from ..interface import ObjectSerializer
from jsonschema_extractor import init_default_extractor
from .converter import create_cattrs_converter
from ...exceptions import SerializationException


class CattrsSerializer(ObjectSerializer):
    """
    A serializer that's intended to become
    the pattern for 2.0
    """

    def __init__(self):
        self._schema_extractor = init_default_extractor()
        self._cattrs_converter = create_cattrs_converter()

    def can_handle(self, cls):
        """
        this will theoretically be compatible with everything,
        as cattrs can handle many basic types as well.
        """
        # cattrs uses a Singledispatch like function
        # under the hood.
        f = self._cattrs_converter._structure_func.dispatch(cls)
        return f != self._cattrs_converter._structure_default

    def load(self, model, value):
        """
        Converts unstructured data into structured data, recursively.
        """
        try:
            return self._cattrs_converter.structure(value, model)
        except (ValueError, TypeError) as e:
            raise SerializationException(str(e))

    def dump(self, model, value):
        """
        Convert attrs data into unstructured data with basic types, recursively:

        - attrs classes => dictionaries
        - Enumeration => values
        - Other types are let through without conversion,
          such as, int, boolean, dict, other classes.
        """
        try:
            return self._cattrs_converter.unstructure(value)
        except (ValueError, TypeError) as e:
            raise SerializationException(str(e))

    def to_json_schema(self, model):
        return self._schema_extractor.extract(model)



from .compound_serializer import ListSerializer
from .cattrs_serializer import CattrsSerializer
from .schematics_serializer import SchematicsSerializer
from .interface import ObjectSerializer
from .serializer_set import ObjectSerializerSet
from .primitive_serializer import (
    NoneSerializer,
    IntSerializer,
    FloatSerializer,
    StringSerializer,
    BoolSerializer,
    DateTimeSerializer,
    DecimalSerializer,
)


DEFAULT_SERIALIZER_LIST = [
    SchematicsSerializer(),
    NoneSerializer(),
    BoolSerializer(),
    StringSerializer(),
    IntSerializer(),
    FloatSerializer(),
    DateTimeSerializer(),
    DecimalSerializer(),
    CattrsSerializer(),
]


def get_default_object_serializer_set():
    s = ObjectSerializerSet(DEFAULT_SERIALIZER_LIST)
    s.serializers.insert(0, ListSerializer(s))
    return s



# from .cattrs_extended_converter import ExtendedConverter
from cattr import Converter
from datetime import datetime
from ...compat import string_type
from schematics.models import Model
from schematics.types import DateTimeType, BaseType
from schematics.exceptions import BaseError


def create_cattrs_converter():
    converter = Converter()
    converter.register_structure_hook(bool, _structure_bool)
    converter.register_structure_hook(string_type, _structure_string)
    converter.register_structure_hook(Model, _structure_schematics)
    converter.register_structure_hook(BaseType, _structure_basetype)
    converter.register_structure_hook(datetime, _structure_datetime)
    converter.register_unstructure_hook(Model, _unstructure_schematics)
    converter.register_unstructure_hook(datetime, _unstructure_datetime)
    converter.register_unstructure_hook(BaseType, _unstructure_basetype)
    return converter


def _structure_bool(data, cls):
    if isinstance(data, bool):
        return data
    return str(data).lower().startswith("t")


def _structure_string(data, cls):
    return str(data)


def _structure_schematics(data, cls):
    value = cls(data)
    try:
        value.validate()
        return value
    except BaseError as de:
        raise ValueError(str(de))


def _unstructure_schematics(data):
    return data.to_primitive()


datetime_type = DateTimeType()


def _structure_datetime(data, cls):
    if not data:
        raise ValueError("datetime is empty")
    return datetime_type.to_native(data)


def _unstructure_datetime(data):
    return data.isoformat()


def _structure_basetype(data, cls):
    return data


def _unstructure_basetype(data, cls):
    return cls().to_primitive(data)



from .decorators import describe, annotate
from .object_serializers import SchematicsSerializer, get_default_object_serializer_set
from .contenttype_serializers import get_default_serializer_set
from .function import TransmuteFunction, Response
from .exceptions import *
from .context import TransmuteContext, default_context
from .handler import process_result
from .param_extractor import NoArgument, ParamExtractor
from .swagger import generate_swagger_html, get_swagger_static_root, SwaggerSpec
from .response_shape import ResponseShape, ResponseShapeComplex, ResponseShapeSimple



from ..compat import all_string_types, string_type
from ..exceptions import SerializationException
from schematics.exceptions import BaseError
from schematics.types import DecimalType, DateTimeType
from decimal import Decimal
from datetime import datetime


class IntSerializer:
    def can_handle(self, cls):
        return issubclass(cls, int)

    @staticmethod
    def to_json_schema(cls):
        return {"type": "integer"}

    @staticmethod
    def load(cls, obj):
        try:
            return int(obj)
        except ValueError as e:
            raise SerializationException(str(e))

    @staticmethod
    def dump(cls, obj):
        return obj


class FloatSerializer:
    def can_handle(self, cls):
        return issubclass(cls, float)

    @staticmethod
    def to_json_schema(cls):
        return {"type": "number"}

    @staticmethod
    def load(cls, obj):
        try:
            return float(obj)
        except ValueError as e:
            raise SerializationException(str(e))

    @staticmethod
    def dump(cls, obj):
        return obj


class StringSerializer:
    def can_handle(self, cls):
        return any(issubclass(cls, t) for t in all_string_types)

    @staticmethod
    def to_json_schema(cls):
        return {"type": "string"}

    @staticmethod
    def load(cls, obj):
        return obj

    @staticmethod
    def dump(cls, obj):
        return obj


class BoolSerializer:
    def can_handle(self, cls):
        return issubclass(cls, bool)

    @staticmethod
    def to_json_schema(cls):
        return {"type": "boolean"}

    @staticmethod
    def load(cls, obj):
        if isinstance(obj, string_type):
            return obj.lower().startswith("t")
        return bool(obj)

    @staticmethod
    def dump(cls, obj):
        return obj


class NoneSerializer(object):
    def can_handle(self, cls):
        return cls is None

    @staticmethod
    def to_json_schema(cls):
        return {"type": "object"}

    @staticmethod
    def load(cls, obj):
        return obj

    @staticmethod
    def dump(cls, obj):
        return obj


class DecimalSerializer(object):

    SERIALIZER = DecimalType()

    def can_handle(self, cls):
        return issubclass(cls, Decimal)

    @staticmethod
    def to_json_schema(cls):
        return {"type": "number"}

    def load(self, cls, obj):
        try:
            return self.SERIALIZER.to_native(obj)
        except BaseError as e:
            raise SerializationException(str(e))

    def dump(self, cls, obj):
        try:
            return self.SERIALIZER.to_primitive(obj)
        except BaseError as e:
            raise SerializationException(str(e))


class DateTimeSerializer(object):

    SERIALIZER = DateTimeType()

    def can_handle(self, cls):
        return issubclass(cls, datetime)

    @staticmethod
    def to_json_schema(cls):
        return {"type": "string", "format": "date-time"}

    def load(self, cls, obj):
        try:
            return self.SERIALIZER.to_native(obj)
        except BaseError as e:
            raise SerializationException(str(e))

    def dump(self, cls, obj):
        try:
            return self.SERIALIZER.to_primitive(obj)
        except BaseError as e:
            raise SerializationException(str(e))



import yaml
from .interface import ContentTypeSerializer
from ..exceptions import SerializationException


class YamlSerializer(ContentTypeSerializer):

    content_type = ["application/x-yaml"]

    @staticmethod
    def dump(data):
        """
        should return back a bytes (or string in python 2),
        representation of your object, to be used in e.g. response
        bodies.
        """
        return yaml.dump(data, default_flow_style=False).encode("UTF-8")

    @classmethod
    def main_type(cls):
        return cls.content_type[0]

    @staticmethod
    def load(raw_bytes):
        """
        given a bytes object, should return a base python data
        structure that represents the object.
        """
        try:
            return yaml.load(raw_bytes)
        except yaml.scanner.ScannerError as e:
            raise SerializationException(str(e))

    @staticmethod
    def can_handle(content_type_name):
        """
        given a content type, returns true if this serializer
        can convert bodies of the given type.
        """
        return "yaml" in content_type_name



import pytest
from transmute_core.object_serializers import ListSerializer
from typing import List


def test_list_serializer(object_serializer_set):
    x = ["10"]

    obj = object_serializer_set.load([int], x)
    assert obj == [10]
    obj = object_serializer_set.dump([int], obj)
    assert obj == [10]


def test_list_serializer_json_schema(object_serializer_set):
    schema = object_serializer_set.to_json_schema([str])
    assert schema == {"type": "array", "items": {"type": "string"}}


def test_object_serializer_set_can_handle(object_serializer_set):
    schema = object_serializer_set.to_json_schema([str])
    assert schema == {"type": "array", "items": {"type": "string"}}


@pytest.mark.parametrize(
    "cls,inp,out,expected_out",
    [(List[int], [1, 2], [1, 2], {"type": "array", "items": {"type": "integer"}})],
)
def test_serializer_values(object_serializer_set, cls, inp, out, expected_out):
    assert object_serializer_set.to_json_schema(cls) == expected_out
    assert object_serializer_set.load(cls, out) == inp
    assert object_serializer_set.dump(cls, inp) == out



import inspect
import pytest

import transmute_core
from transmute_core.function import TransmuteFunction
from transmute_core.http_parameters import Param
from transmute_core.function.signature import Argument
from .conftest import Pet


@transmute_core.describe(paths="/", tags=["example"])
@transmute_core.annotate({"return": int})
def raw_func():
    """
    foo bar

    baz
    """
    return 12345


@pytest.fixture
def func():
    return TransmuteFunction(raw_func)


def test_callable(func):
    """ test function is callable, and routes to the inner function. """
    assert func() == 12345


def test_description(func):
    """ description should be the docstring. """
    assert func.description == inspect.cleandoc(raw_func.__doc__)


def test_summary(func):
    """ summary should be the first non-empty line in the docstring. """
    assert func.summary == "foo bar"


def test_tags(func):
    assert func.tags == set(["example"])


def test_return_type(func):
    """ test function is callable, and routes to the inner function. """
    assert func.get_response_by_code(200) is int


def test_swagger_schema_has_object(func):
    swagger = func.get_swagger_operation()
    assert swagger.responses["200"].schema.to_primitive() == {"type": "integer"}


def test_swagger_operation_has_operation_id(func):
    op = func.get_swagger_operation()
    assert op.operationId == "raw_func"


def test_swagger_operation_has_tags(func):
    op = func.get_swagger_operation()
    assert op.tags == ["example"]


def test_swagger_schema_path(func):
    swagger = func.get_swagger_path()
    assert swagger.get_.responses["200"].schema.to_primitive() == {"type": "integer"}


def test_body_param_func(single_body_transmute_func):
    func = single_body_transmute_func
    assert isinstance(func.parameters.body, Param)
    assert func.parameters.body.argument_name == "body"


def test_function_raises_exception_on_path_missing():
    @transmute_core.describe(paths=[])
    def func():
        pass

    with pytest.raises(transmute_core.InvalidTransmuteDefinition):
        TransmuteFunction(func)



import pytest
from transmute_core import (
    annotate,
    describe,
    default_context,
    get_default_serializer_set,
    get_default_object_serializer_set,
    Response,
    SchematicsSerializer,
    TransmuteFunction,
)
from schematics.models import Model
from schematics.types import StringType, IntType


@pytest.fixture
def context():
    return default_context


@pytest.fixture
def serializer_set():
    return get_default_serializer_set()


@pytest.fixture
def object_serializers():
    return get_default_object_serializer_set()


@pytest.fixture
def object_serializer_set():
    return get_default_object_serializer_set()


@pytest.fixture
def serializer():
    return SchematicsSerializer()


@pytest.fixture
def transmute_func():
    @describe(paths="/api/v1/multiply")
    @annotate({"left": int, "right": int, "return": int})
    def multiply(left, right):
        return left * right

    return TransmuteFunction(multiply)


@pytest.fixture
def transmute_func_custom_code():
    @describe(paths="/api/v1/multiply", success_code=201)
    @annotate({"left": int, "right": int, "return": int})
    def multiply(left, right):
        return left * right

    return TransmuteFunction(multiply)


@pytest.fixture
def transmute_func_post():
    @describe(paths="/api/v1/multiply", methods=["POST"])
    @annotate({"left": int, "right": int, "return": int})
    def multiply(left, right):
        return left * right

    return TransmuteFunction(multiply)


class Pet(Model):

    kind = StringType(required=True)
    age = IntType()


@pytest.fixture
def single_body_transmute_func():
    @describe(paths="/", body_parameters="body")
    @annotate({"body": Pet})
    def body_param_func():
        return None

    return TransmuteFunction(body_param_func)


@pytest.fixture
def complex_transmute_func():
    @describe(paths="/api/v1/adopt")
    @annotate({"return": Pet})
    def adopt():
        return Pet({"kind": "dog", "age": 5})

    return TransmuteFunction(adopt)


@pytest.fixture
def response_transmute_func():
    @describe(
        paths="/api/v1/create_if_authorized/",
        response_types={
            401: {"type": str, "description": "unauthorized"},
            201: {"type": bool, "headers": {"location": {"type": str}}},
        },
    )
    @annotate({"username": str})
    def create_if_authorized(username):
        if username != "im the boss":
            return Response("this is unauthorized!", 201)
        else:
            return Response(True, 401)

    return TransmuteFunction(create_if_authorized)



import pytest
from transmute_core.compat import string_type
from transmute_core.exceptions import SerializationException
from datetime import datetime

NOW = datetime.now()
UTCNOW = datetime.utcnow()


@pytest.mark.parametrize("inp, expected_output", [("10", 10), ("-1", -1)])
def test_int_load_happy(object_serializer_set, inp, expected_output):
    """ test all happy cases for the integer serializer """
    assert object_serializer_set.load(int, inp) == expected_output


@pytest.mark.parametrize("unhappy_input", ["foo", "bar"])
def test_int_load_unhappy(object_serializer_set, unhappy_input):
    """ test all unhappy cases for the integer serializer """
    with pytest.raises(SerializationException):
        object_serializer_set.load(int, unhappy_input)


@pytest.mark.parametrize("inp, expected_output", [("10", 10), ("1.0", 1.0)])
def test_float_load_happy(object_serializer_set, inp, expected_output):
    """ test all happy cases for the integer serializer """
    assert object_serializer_set.load(float, inp) == expected_output


@pytest.mark.parametrize("unhappy_input", ["foo", "bar"])
def test_float_load_unhappy(object_serializer_set, unhappy_input):
    """ test all unhappy cases for the integer serializer """
    with pytest.raises(SerializationException):
        object_serializer_set.load(float, unhappy_input)


@pytest.mark.parametrize(
    "inp, expected_output",
    [("true", True), ("false", False), (True, True), (False, False)],
)
def test_bool_load_happy(object_serializer_set, inp, expected_output):
    assert object_serializer_set.load(bool, inp) is expected_output


@pytest.mark.parametrize("inp, expected_output", [("foo", "foo")])
def test_string_load_happy(object_serializer_set, inp, expected_output):
    assert object_serializer_set.load(string_type, inp) == expected_output


@pytest.mark.parametrize("inp, expected_output", [(NOW.isoformat(), NOW)])
def test_datetime_load_happy(object_serializer_set, inp, expected_output):
    assert object_serializer_set.load(datetime, inp) == expected_output


@pytest.mark.parametrize("inp, expected_output", [(NOW, NOW.isoformat())])
def test_datetime_dump_unhappy(object_serializer_set, inp, expected_output):
    assert object_serializer_set.dump(datetime, inp) == expected_output


@pytest.mark.parametrize("inp", [("")])
def test_datetime_load_unhappy(object_serializer_set, inp):
    with pytest.raises(SerializationException):
        object_serializer_set.load(datetime, inp)


@pytest.mark.parametrize("inp, out", [(None, None), ("s", "s")])
def test_none_serialzer(object_serializer_set, inp, out):
    assert object_serializer_set.load(None, inp) == out



import pytest
import swagger_schema
from transmute_core import default_context, describe, annotate, TransmuteFunction
from transmute_core.swagger import (
    generate_swagger_html,
    get_swagger_static_root,
    SwaggerSpec,
)


def test_generate_swagger():
    swagger_static_root = "/statics/_swagger"
    swagger_json_url = "/swagger.json"
    body = generate_swagger_html(swagger_static_root, swagger_json_url)
    assert swagger_static_root in body
    assert swagger_json_url in body


def test_get_swagger_static_root():
    assert "static" in get_swagger_static_root()


@pytest.mark.parametrize(
    ["swagger_definition_kwargs", "expected_info"],
    [
        (
            # Use default title and version
            {},
            {"title": "example", "version": "1.0"},
        ),
        (
            # Use default version
            {"title": "title-only app"},
            {"title": "title-only app", "version": "1.0"},
        ),
        (
            {"title": "test app", "description": "unit test app", "version": "1.1.38"},
            {"title": "test app", "description": "unit test app", "version": "1.1.38"},
        ),
    ],
)
def test_swagger_definition_generation(swagger_definition_kwargs, expected_info):
    """
    swagger routes should be able to generate a proper
    spec.
    """
    routes = SwaggerSpec()
    assert routes.swagger_definition(**swagger_definition_kwargs) == {
        "info": expected_info,
        "paths": {},
        "swagger": "2.0",
    }


def test_swagger_transmute_definition_info(transmute_func):
    """
    swagger routes should be able to generate a proper
    spec.
    """
    routes = SwaggerSpec()
    routes.add_func(transmute_func, default_context)
    assert routes.swagger_definition() == {
        "info": {"title": "example", "version": "1.0"},
        "paths": {
            "/api/v1/multiply": transmute_func.get_swagger_path(
                default_context
            ).to_primitive()
        },
        "swagger": "2.0",
    }


def test_multiple_response_types(response_transmute_func):
    """ multiple response types should be indicated as such in
    the swagger documentation.
    """
    routes = SwaggerSpec()
    routes.add_func(response_transmute_func, default_context)
    definition = routes.swagger_definition()
    path = definition["paths"]["/api/v1/create_if_authorized/"]
    responses = path["get"]["responses"]
    assert (
        responses["201"]
        == swagger_schema.Response(
            {
                "description": "",
                "schema": {"type": "boolean"},
                "headers": {"location": {"type": "string"}},
            }
        ).to_primitive()
    )
    assert (
        responses["401"]
        == swagger_schema.Response(
            {"description": "unauthorized", "schema": {"type": "string"}}
        ).to_primitive()
    )
    assert "200" not in responses
    assert "400" in responses


def test_swagger_add_path(transmute_func):
    """
    add_path should add the specified path to the main swagger object.
    """
    routes = SwaggerSpec()
    swagger_path = transmute_func.get_swagger_path(default_context)
    for p in transmute_func.paths:
        routes.add_path(p, swagger_path)
    assert routes.swagger_definition() == {
        "info": {"title": "example", "version": "1.0"},
        "paths": {
            "/api/v1/multiply": transmute_func.get_swagger_path(
                default_context
            ).to_primitive()
        },
        "swagger": "2.0",
    }


def test_basepath_override():
    assert SwaggerSpec().swagger_definition(base_path="dummy_path") == {
        "basePath": "dummy_path",
        "info": {"title": "example", "version": "1.0"},
        "paths": {},
        "swagger": "2.0",
    }


def test_swagger_get_post(transmute_func, transmute_func_post):
    """
    adding different paths of diffrent methods should have both
    present in the spec.
    """
    routes = SwaggerSpec()
    routes.add_func(transmute_func, default_context)
    routes.add_func(transmute_func_post, default_context)
    spec = routes.swagger_definition()
    assert "get" in spec["paths"]["/api/v1/multiply"]
    assert "post" in spec["paths"]["/api/v1/multiply"]


def test_swagger_parameter_description():
    """
    if parameter descriptions are added to a function, they
    should appear in the swagger json.
    """
    parameter_descriptions = {
        "left": "the left operand",
        "right": "the right operand",
        "header": "the header",
        "path": "the path",
        "return": "the result",
    }

    @describe(
        paths="/api/v1/adopt/{path}",
        parameter_descriptions=parameter_descriptions,
        header_parameters=["header"],
    )
    @annotate({"left": int, "right": int, "header": str, "path": str, "return": int})
    def adopt(left, right, header, path):
        return left + right

    func = TransmuteFunction(adopt)

    routes = SwaggerSpec()
    routes.add_func(func, default_context)
    spec = routes.swagger_definition()

    params = spec["paths"]["/api/v1/adopt/{path}"]["get"]["parameters"]
    for param in spec["paths"]["/api/v1/adopt/{path}"]["get"]["parameters"]:
        assert parameter_descriptions[param["name"]] == param["description"]
    assert (
        parameter_descriptions["return"]
        == spec["paths"]["/api/v1/adopt/{path}"]["get"]["responses"]["200"]["schema"][
            "description"
        ]
    )



import attr
import cProfile
import attr
import json
import pytest
import sys
from attr.validators import instance_of
from transmute_core import (
    TransmuteFunction,
    describe,
    annotate,
    ParamExtractor,
    NoArgument,
)
from schematics.models import Model
from schematics.types import StringType, BooleanType, IntType
from schematics.types.compound import ModelType
from .utils import execute


class User(Model):
    name = StringType()
    age = IntType()


class ComplexModel(Model):
    user = ModelType(User)
    description = StringType()
    is_allowed = BooleanType()


@describe(paths="/foo", body_parameters="body")
@annotate({"body": ComplexModel, "return": ComplexModel})
def complex_body_method(body):
    return body


@describe(paths="/foo", body_parameters="body")
@annotate({"body": int, "return": int})
def simple_body_method(body):
    return body


@describe(paths="/foo", body_parameters="body")
@annotate({"body": str, "return": str})
def body_string(body):
    return body


def test_large_str_benchmark(benchmark, context):
    """
    a benchmark of a fake full execution flow of a transmute function.
    """
    s = "a" * 100000

    func = TransmuteFunction(body_string)
    obj_json = json.dumps(s)

    benchmark(lambda: execute(context, func, obj_json))


def test_complex_benchmark(benchmark, context):
    """
    a benchmark of a fake full execution flow of a transmute function.
    """
    obj = ComplexModel(
        {
            "user": {"name": "Richard Stallman", "age": 104},
            "description": "this is a test",
            "is_allowed": True,
        }
    )

    complex_func = TransmuteFunction(complex_body_method)
    complex_json = json.dumps(context.serializers.dump(type(obj), obj))

    benchmark(lambda: execute(context, complex_func, complex_json))


def test_simple_benchmark(benchmark, context):

    simple_func = TransmuteFunction(simple_body_method)
    simple_json = json.dumps(1)

    benchmark(lambda: execute(context, simple_func, simple_json))



from ...object_serializers.primitive_serializer import BoolSerializer
import pytest


@pytest.mark.parametrize(
    "object, result",
    [
        ("true", True),
        ("false", False),
        ("some_other_str", False),
        ([], False),
        (["non-empty-list"], True),
        ({}, False),
        ({"non_empty_dict": "test_val"}, True),
    ],
)
def test_primitive_serializer_bool_serializer(object, result):
    serializer = BoolSerializer()
    assert BoolSerializer.load(serializer, object) == result



import pytest
from transmute_core.http_parameters import (
    get_swagger_parameters,
    ParamSet,
    Param,
    Parameters,
)
from transmute_core.function.signature import Argument, NoDefault
from datetime import datetime


def test_post_schema_swagger(parameter_post_schema, context):
    assert get_swagger_parameters(parameter_post_schema, context)[0].to_primitive() == {
        "in": "body",
        "name": "body",
        "required": True,
        "description": "",
        "schema": {"type": "integer"},
    }


def test_header_only_schema(context):
    params = Parameters(
        path=ParamSet(
            {
                "path": Param(
                    argument_name="path", arginfo=Argument("path", NoDefault, int)
                )
            }
        )
    )
    assert get_swagger_parameters(params, context)[0].to_primitive() == {
        "in": "path",
        "name": "path",
        "required": True,
        "type": "integer",
        "description": "",
    }


@pytest.mark.parametrize(
    "inp, expected",
    [
        (
            Parameters(
                query=ParamSet(
                    {
                        "query": Param(
                            argument_name="query",
                            arginfo=Argument("query", NoDefault, datetime),
                        )
                    }
                )
            ),
            {
                "in": "query",
                "name": "query",
                "required": True,
                "type": "string",
                "format": "date-time",
                "collectionFormat": "multi",
                "description": "",
            },
        ),
        (
            Parameters(
                path=ParamSet(
                    {
                        "path": Param(
                            argument_name="path",
                            arginfo=Argument("path", NoDefault, datetime),
                        )
                    }
                )
            ),
            {
                "in": "path",
                "name": "path",
                "required": True,
                "type": "string",
                "format": "date-time",
                "description": "",
            },
        ),
        (
            Parameters(
                header=ParamSet(
                    {
                        "path": Param(
                            argument_name="path",
                            arginfo=Argument("path", NoDefault, datetime),
                        )
                    }
                )
            ),
            {
                "in": "header",
                "name": "path",
                "required": True,
                "type": "string",
                "format": "date-time",
                "description": "",
            },
        ),
    ],
)
def test_additional_type_info(context, inp, expected):
    """
    query parameter should have additional type data
    (like format) added to it.
    """
    assert get_swagger_parameters(inp, context)[0].to_primitive() == expected



from .interface import ContentTypeSerializer
from .json_serializer import JsonSerializer
from .yaml_serializer import YamlSerializer
from .serializer_set import SerializerSet, NoSerializerFound


def get_default_serializer_set():
    return SerializerSet([JsonSerializer(), YamlSerializer()])



import attr


class NoDefault(object):
    def __str__(self):
        return "NoDefault"

    def __repr__(self):
        return "NoDefault"


NoDefault = NoDefault()


@attr.s
class Argument(object):
    name = attr.ib()
    default = attr.ib()
    type = attr.ib()


class FunctionSignature(object):

    NoDefault = NoDefault

    def __init__(self, args, kwargs):
        self.args = args
        self.kwargs = kwargs

    def get_argument(self, key):
        if key in self.kwargs:
            return self.kwargs[key]
        for arg in self.args:
            if arg.name == key:
                return arg

    @staticmethod
    def from_argspec(argspec):
        """
        retrieve a FunctionSignature object
        from the argspec and the annotations passed.
        """
        attributes = getattr(argspec, "args", []) + getattr(argspec, "keywords", [])
        defaults = argspec.defaults or []

        arguments, keywords = [], {}

        attribute_list = (
            attributes[: -len(defaults)] if len(defaults) != 0 else attributes[:]
        )
        for name in attribute_list:
            if name == "self":
                continue
            typ = argspec.annotations.get(name)
            arguments.append(Argument(name, NoDefault, typ))

        if len(defaults) != 0:
            for name, default in zip(attributes[-len(defaults) :], defaults):
                typ = argspec.annotations.get(name)
                keywords[name] = Argument(name, default, typ)

        return FunctionSignature(arguments, keywords)

    def __iter__(self):
        for arg in self.args:
            yield arg
        for kwarg in self.kwargs.values():
            yield kwarg

    def split_args(self, arg_dict):
        """
        given a dictionary of arguments, split them into
        args and kwargs

        note: this destroys the arg_dict passed. if you need it,
        create a copy first.
        """
        pos_args = []
        for arg in self.args:
            pos_args.append(arg_dict[arg.name])
            del arg_dict[arg.name]
        return pos_args, arg_dict



import sys
import inspect
from collections import namedtuple

is_py3 = sys.version_info[0] >= 3

FullArgSpec = namedtuple(
    "FullArgSpec",
    [
        "args",
        "varargs",
        "varkw",
        "defaults",
        "kwonlyargs",
        "kwonlydefaults",
        "annotations",
    ],
)


def getfullargspec(func):
    if is_py3:
        return inspect.getfullargspec(func)
    else:
        argspec = inspect.getargspec(func)
        annotations = getattr(func, "__annotations__", {})
        return FullArgSpec(
            argspec.args,
            argspec.varargs,
            argspec.keywords,
            argspec.defaults,
            [],
            None,
            annotations,
        )


if is_py3:
    string_type = str  # noqa
    all_string_types = [str]
else:
    string_type = basestring  # noqa
    all_string_types = [basestring, str, unicode]



import uuid
import pytest
from datetime import datetime
from schematics.models import Model
from schematics.types import (
    StringType,
    IntType,
    BaseType,
    UUIDType,
    URLType,
    Serializable,
    UTCDateTimeType,
)
from schematics.types.compound import DictType, ModelType
from schematics.exceptions import ValidationError
from transmute_core.exceptions import SerializationException, NoSerializerFound


class Card(Model):
    name = StringType()
    price = IntType()


def greater_than_zero(value):
    if value < 0:
        raise ValidationError("must be greater than zero")
    return value


class Person(Model):
    age = IntType(required=True, validators=[greater_than_zero])
    bio = StringType(max_length=5)


card_dict = {"name": "foo", "price": 10}
card = Card(card_dict)


@pytest.mark.parametrize(
    "typ,inp,out", [(Card, card, card_dict), ([Card], [card], [card_dict])]
)
def test_schematics_integration_dump(object_serializer_set, typ, inp, out):
    assert object_serializer_set.dump(typ, inp) == out


@pytest.mark.parametrize(
    "typ, bad_data",
    [
        (Person, {"age": -1, "bio": "foo"}),
        (Person, {"age": 20, "bio": "fooooooooo"}),
        (StringType(max_length=4), "foooooo"),
    ],
)
def test_schematics_validate_is_called(object_serializer_set, typ, bad_data):
    with pytest.raises(SerializationException):
        object_serializer_set.load(typ, bad_data)


now = datetime.now()


@pytest.mark.parametrize(
    "typ,inp,out",
    [(Card, card_dict, card), ([Card], [card_dict], [card]), (datetime, now, now)],
)
def test_schematics_integration_load(object_serializer_set, typ, inp, out):
    assert object_serializer_set.load(typ, inp) == out


def test_schematics_to_json_schema(object_serializer_set):
    assert object_serializer_set.to_json_schema(Card) == {
        "properties": {"name": {"type": "string"}, "price": {"type": "integer"}},
        "title": "Card",
        "type": "object",
    }


@pytest.mark.parametrize(
    "inp, expected",
    [
        (UUIDType, {"type": "string", "format": "uuid"}),
        (URLType, {"type": "string", "format": "url"}),
        (StringType, {"type": "string"}),
        (BaseType, {"type": "object"}),
        (IntType, {"type": "integer"}),
        (datetime, {"type": "string", "format": "date-time"}),
        (UTCDateTimeType, {"type": "string", "format": "date-time"}),
        (Serializable(fget=lambda: None,
                      type=StringType()), {"type": "string"}),
    ],
)
def test_to_json_schema(object_serializer_set, inp, expected):
    assert object_serializer_set.to_json_schema(inp) == expected


def test_schematics_to_json_schema_required_value(object_serializer_set):
    class CardNameRequired(Model):
        name = StringType(required=True)
        price = IntType()

    assert object_serializer_set.to_json_schema(CardNameRequired) == {
        "properties": {"name": {"type": "string"}, "price": {"type": "integer"}},
        "title": "CardNameRequired",
        "required": ["name"],
        "type": "object",
    }


def test_type_not_found_defaults_to_object(object_serializer_set):
    class MyType(BaseType):
        pass

    assert object_serializer_set.to_json_schema(MyType()) == {"type": "object"}


def test_non_model_or_primitive_raises_exception(object_serializer_set):
    class MyType(object):
        pass

    with pytest.raises(NoSerializerFound):
        object_serializer_set.to_json_schema(MyType)


def test_to_json_list(object_serializer_set):
    assert object_serializer_set.to_json_schema([int]) == {
        "type": "array",
        "items": {"type": "integer"},
    }


def test_to_json_dict(object_serializer_set):
    assert object_serializer_set.to_json_schema(DictType(IntType())) == {
        "type": "object",
        "additionalProperties": {"type": "integer"},
    }


def test_serialize_type(object_serializer_set):
    """ object_serializer_set should be able to serialize a raw type. """
    assert object_serializer_set.dump(DictType(StringType()), {"key": "foo"}) == {
        "key": "foo"
    }


def test_dict_type_object_serializer_set(object_serializer_set):
    """ serializer should be able to serialize a raw type. """
    assert object_serializer_set.dump(DictType(StringType()), {"key": "foo"}) == {
        "key": "foo"
    }


def test_uuid_serializer(object_serializer_set):
    i = uuid.uuid4()
    dumped = object_serializer_set.dump(UUIDType, i)
    assert dumped == str(i)
    assert object_serializer_set.load(UUIDType, dumped) == i


@pytest.mark.parametrize(
    "cls, should_handle",
    [(int, True), (float, True), (datetime, True), (ModelType(Card), True)],
)
def test_can_handle(object_serializer_set, cls, should_handle):
    """
    the object serializer set should not raise an exception if it can handle these types.
    """
    object_serializer_set[cls]



import attr
import cattr
import json
import pytest
import sys
import mock

from attr.validators import instance_of
from datetime import datetime
from mock import Mock
from transmute_core.exceptions import SerializationException
from typing import List
from transmute_core.object_serializers.cattrs_serializer import CattrsSerializer
from transmute_core import describe, annotate, TransmuteFunction
from .utils import execute


@pytest.fixture
def cattrs_serializer():
    return CattrsSerializer()


class Player(object):
    def __init__(self, name, score):
        self.name = name
        self.score = score


@attr.s
class Card(object):
    name = attr.ib(type=str, default="")
    price = attr.ib(type=float, default=1.0)


@attr.s
class Hand(object):
    cards = attr.ib(type=List[Card])


@attr.s
class Person(object):
    age = attr.ib(type=int)
    bio = attr.ib(type=str, default="")

    @age.validator
    def greater_than_zero(self, attribute, value):
        if value < 0:
            raise ValueError("must be greater than zero")
        return value


card_dict = {"name": "foo", "price": 10}
card = Card(**card_dict)
player = Player("kobe", 81)


@pytest.mark.parametrize(
    "typ,inp,out",
    [
        (Card, card, card_dict),
        (List[Card], [card], [card_dict]),
        (Player, player, player),
    ],
)
def test_attrs_integration_dump(cattrs_serializer, typ, inp, out):
    assert cattrs_serializer.dump(typ, inp) == out


def test_attrs_integration_dump_exception(cattrs_serializer):
    def mock_return(inp):
        raise ValueError("Random Exception")

    cattrs_serializer._cattrs_converter = Mock()
    cattrs_serializer._cattrs_converter.unstructure = mock_return
    with pytest.raises(SerializationException):
        assert cattrs_serializer.dump(str, "random_str")


@pytest.mark.parametrize("typ,inp", [([Card], [card_dict])])
def test_attrs_integration_load_exception(cattrs_serializer, typ, inp):
    with pytest.raises(SerializationException):
        cattrs_serializer.load(typ, inp)


def test_attrs_validate_is_called(cattrs_serializer):
    with pytest.raises(SerializationException):
        cattrs_serializer.load(Person, {"age": -1, "bio": "foo"})


@pytest.mark.parametrize(
    "typ,inp,out", [(Card, card_dict, card), (List[Card], [card_dict], [card])]
)
def test_attrs_integration_load(cattrs_serializer, typ, inp, out):
    assert cattrs_serializer.load(typ, inp) == out


@attr.s
class Employee(object):
    name = attr.ib(type=str)
    birthday = attr.ib(type=datetime)


birthday = datetime.utcnow()
employee_dict = {"name": "Kobe", "birthday": birthday.isoformat()}
employee = Employee(name="Kobe", birthday=birthday)


@pytest.mark.parametrize("typ,inp,out", [(Employee, employee_dict, employee)])
def test_attrs_integration_load_datetime(cattrs_serializer, typ, inp, out):
    assert cattrs_serializer.load(typ, inp) == out


def test_to_json_schema(cattrs_serializer):
    assert cattrs_serializer.to_json_schema(Card) == {
        "type": "object",
        "title": "Card",
        "properties": {"name": {"type": "string"}, "price": {"type": "number"}},
        "required": [],
    }

    assert cattrs_serializer.to_json_schema(Hand) == {
        "type": "object",
        "title": "Hand",
        "properties": {
            "cards": {
                "type": "array",
                "items": {
                    "type": "object",
                    "title": "Card",
                    "properties": {
                        "name": {"type": "string"},
                        "price": {"type": "number"},
                    },
                    "required": [],
                },
            }
        },
        "required": ["cards"],
    }


@attr.s
class UserAttrs(object):
    name = attr.ib(validator=instance_of(str))
    age = attr.ib(validator=instance_of(int))


@attr.s
class UserAttrs(object):
    name = attr.ib(type=str)
    age = attr.ib(type=int)


@attr.s
class ComplexModelAttrs(object):
    user = attr.ib(type=UserAttrs)
    description = attr.ib(type=str)
    is_allowed = attr.ib(type=bool)


@describe(paths="/foo", body_parameters="body")
@annotate({"body": ComplexModelAttrs, "return": ComplexModelAttrs})
def complex_body_method_attrs(body):
    return body


def test_complex_benchmark_attrs(benchmark, context):
    """
    a benchmark of a fake full execution flow of a transmute function.
    """
    obj = ComplexModelAttrs(
        user=UserAttrs(name="Richard Stallman", age=104),
        description="this is a test",
        is_allowed=True,
    )

    complex_func = TransmuteFunction(complex_body_method_attrs)
    complex_json = json.dumps(context.serializers.dump(type(obj), obj))

    benchmark(lambda: execute(context, complex_func, complex_json))



import pytest
from transmute_core import describe, annotate, Response
from transmute_core.context import TransmuteContext
from transmute_core.response_shape import ResponseShapeComplex


@pytest.fixture
def context():
    return TransmuteContext(response_shape=ResponseShapeComplex)


def test_complex_response_shape_in_func(transmute_func, context):
    swagger_op = transmute_func.get_swagger_operation(context=context)
    swagger_op.responses["200"].description == "success"
    swagger_op.responses["200"].schema.to_primitive() == {
        "title": "SuccessObject",
        "type": "object",
        "properties": {
            "result": context.serializers.to_json_schema(
                transmute_func.response_types[200].type
            ),
            "success": {"type": "boolean"},
            "code": {"type": "number"},
        },
        "required": ["success", "result", "code"],
    }



from .compat import string_type
from .function import TransmuteAttributes


def describe(**kwargs):
    """ describe is a decorator to customize the rest API
    that transmute generates, such as choosing
    certain arguments to be query parameters or
    body parameters, or a different method.

    :param list(str) paths: the path(s) for the handler to represent (using swagger's syntax for a path)

    :param list(str) methods: the methods this function should respond to. if non is set, transmute defaults to a GET.

    :param list(str) query_parameters: the names of arguments that
           should be query parameters. By default, all arguments are query_or path parameters for a GET request.

    :param body_parameters: the names of arguments that should be body parameters.
           By default, all arguments are either body or path parameters for a non-GET request.

           in the case of a single string, the whole body is validated against a single object.

    :type body_parameters: List[str] or str

    :param list(str) header_parameters: the arguments that should be passed into the header.

    :param list(str) path_parameters: the arguments that are specified by the path. By default, arguments
            that are found in the path are used first before the query_parameters and body_parameters.

    :param list(str) parameter_descriptions: descriptions for each parameter, keyed by attribute name.
            this will appear in the swagger documentation.
    """
    # if we have a single method, make it a list.
    if isinstance(kwargs.get("paths"), string_type):
        kwargs["paths"] = [kwargs["paths"]]
    if isinstance(kwargs.get("methods"), string_type):
        kwargs["methods"] = [kwargs["methods"]]
    attrs = TransmuteAttributes(**kwargs)

    def decorator(f):
        if hasattr(f, "transmute"):
            f.transmute = f.transmute | attrs
        else:
            f.transmute = attrs
        return f

    return decorator


def annotate(annotations):
    """
    in python2, native annotions on parameters do not exist:

    .. code-block:: python

        def foo(a : str, b: int) -> bool:
            ...

    this provides a way to provide attribute annotations:

    .. code-block:: python

        @annotate({"a": str, "b": int, "return": bool})
        def foo(a, b):
            ...
    """

    def decorate(func):
        func.__annotations__ = annotations
        return func

    return decorate



from swagger_schema import BodyParameter, QueryParameter, HeaderParameter, PathParameter
from ..function.signature import NoDefault
from .param_set import Param, ParamSet


def get_swagger_parameters(parameters, context):
    ret_parameters = []
    for name, param in parameters.query.items():
        arginfo = param.arginfo
        params = {
            "name": name,
            "description": param.description,
            "required": arginfo.default is NoDefault,
            "collectionFormat": "multi",
        }
        params.update(context.serializers.to_json_schema(arginfo.type))
        ret_parameters.append(QueryParameter(params))

    for name, param in parameters.header.items():
        arginfo = param.arginfo
        params = {
            "name": name,
            "description": param.description,
            "required": arginfo.default is NoDefault,
        }
        params.update(context.serializers.to_json_schema(arginfo.type))
        ret_parameters.append(HeaderParameter(params))

    body_param = _build_body_schema(context.serializers, parameters.body)
    if body_param is not None:
        ret_parameters.append(body_param)

    for name, param in parameters.path.items():
        arginfo = param.arginfo
        params = {
            "name": name,
            "description": param.description,
            "required": arginfo.default is NoDefault,
        }
        params.update(context.serializers.to_json_schema(arginfo.type))
        ret_parameters.append(PathParameter(params))

    return ret_parameters


def _build_body_schema(serializer, body_parameters):
    """ body is built differently, since it's a single argument no matter what. """
    description = ""
    if isinstance(body_parameters, Param):
        schema = serializer.to_json_schema(body_parameters.arginfo.type)
        description = body_parameters.description
        required = True
    else:
        if len(body_parameters) == 0:
            return None
        required = set()
        body_properties = {}
        for name, param in body_parameters.items():
            arginfo = param.arginfo
            body_properties[name] = serializer.to_json_schema(arginfo.type)
            body_properties[name]["description"] = param.description
            if arginfo.default is NoDefault:
                required.add(name)
        schema = {
            "type": "object",
            "required": list(required),
            "properties": body_properties,
        }
        required = len(required) > 0
    return BodyParameter(
        {
            "name": "body",
            "description": description,
            "required": required,
            "schema": schema,
        }
    )



from transmute_core.attributes import TransmuteAttributes, ResponseType
from schematics.types.net import URLType
from swagger_schema import Response


def test_merge():
    left = TransmuteAttributes(methods=["POST"], query_parameters=["a"], tags=["t"])
    right = TransmuteAttributes(
        methods=["PUT"], query_parameters=["b"], body_parameters=["c"]
    )

    joined = left | right

    assert joined.tags == set(["t"])
    assert joined.methods == set(["POST", "PUT"])
    assert joined.query_parameters == set(["a", "b"])
    assert joined.body_parameters == set(["c"])


def test_merge_response_type_by_code():
    left = TransmuteAttributes(
        response_types={200: {"type": bool}, 201: {"type": bool}}
    )
    right = TransmuteAttributes(response_types={201: {"type": str}})
    joined = left | right
    assert joined.response_types == {
        200: ResponseType(type=bool),
        201: ResponseType(type=str),
    }


def test_swagger_definition_if_none(context):
    """ a response type that is not specified, should result in a generic swagger schema. """
    response_type = ResponseType(type=None)
    definition = response_type.swagger_definition(context)
    assert definition.to_primitive() == {
        "description": "",
        "schema": {"type": "object"},
    }


def test_header_in_response():
    left = TransmuteAttributes(
        response_types={200: {"type": bool, "headers": {"location": {"type": URLType}}}}
    )
    assert left.response_types == {
        200: ResponseType(type=bool, headers={"location": {"type": URLType}})
    }


def test_merge_response_success_code():
    left = TransmuteAttributes(success_code=200)
    right = TransmuteAttributes(success_code=201)
    joined = left | right
    assert joined.success_code == 201


def test_merge_body_parameters_argument():
    right = TransmuteAttributes()
    left = TransmuteAttributes(body_parameters="body")
    joined = left | right
    assert joined.body_parameters == "body"


def test_duplicate_tags():
    left = TransmuteAttributes(
        methods=["POST"], query_parameters=["a"], tags=["t", "t"]
    )

    assert left.tags == set(["t"])



from .interface import ObjectSerializer
from collections import OrderedDict
from datetime import datetime
from schematics.types import (
    BaseType,
    BooleanType,
    DecimalType,
    FloatType,
    IntType,
    NumberType,
    StringType,
    Serializable,
    DateTimeType,
    UUIDType,
    URLType,
)
from schematics.models import ModelMeta, Model
from schematics.types.compound import ListType, ModelType, DictType
from schematics.exceptions import BaseError, ConversionError
from schematics.transforms import get_import_context
from ..exceptions import SerializationException
from decimal import Decimal
from ..compat import all_string_types

JSON_SCHEMA_MAP = OrderedDict(
    [
        (BooleanType, {"type": "boolean"}),
        (IntType, {"type": "integer"}),
        (NumberType, {"type": "number"}),
        (UUIDType, {"type": "string", "format": "uuid"}),
        (URLType, {"type": "string", "format": "url"}),
        (StringType, {"type": "string"}),
        (DateTimeType, {"type": "string", "format": "date-time"}),
        (BaseType, {"type": "object"}),
    ]
)


class SchematicsSerializer(ObjectSerializer):
    """
    An ObjectSerializer which allows the serialization of
    basic types and schematics models.

    The valid types that SchematicsSerializer supports are:

    - int
    - float
    - bool
    - decimal
    - string
    - none
    - lists, in the form of [Type] (e.g. [str])
    - any type that extends the schematics.models.Model.
    """

    VALID_BASE_CLASSES = [BaseType, ModelMeta, Model, Serializable]

    def can_handle(self, cls):
        if cls in self._models:
            return True
        if any(isinstance(cls, t) for t in self.VALID_BASE_CLASSES):
            return True
        if not isinstance(cls, type):
            return False
        if any(issubclass(cls, t) for t in self.VALID_BASE_CLASSES):
            return True

    def __init__(self, builtin_models=None):
        self._models = {}

    def _translate_to_model(self, model):
        model = self._to_key(model)

        if isinstance(model, BaseType):
            return model

        if isinstance(model, ModelMeta):
            return ModelType(model)

        if model not in self._models and isinstance(model, tuple):
            self._models[model] = ListType(self._translate_to_model(model[0]))

        if model in self._models:
            return self._models[model]

        return model

    @staticmethod
    def _to_key(model):
        if isinstance(model, list):
            return tuple(model)
        return model

    def load(self, model, value):
        model = _enforce_instance(model)
        try:
            context = get_import_context(oo=True)
            model = self._translate_to_model(model)
            result = model(value, context=context)
            if isinstance(model, ModelType) or not isinstance(model, BaseType):
                result.validate()
            else:
                model.validate(result, context=context)
            return result
        except BaseError as e:
            raise SerializationException(str(e))
        except ConversionError as e:
            raise SerializationException(str(e))

    def dump(self, model, value):
        model = _enforce_instance(model)
        try:
            model = self._translate_to_model(model)
            return model.to_primitive(value)
        except BaseError as e:
            raise SerializationException(str(e))

    def to_json_schema(self, model):
        if isinstance(model, Serializable):
            model = model.type
        model = _enforce_instance(model)
        model = self._translate_to_model(model)
        return _to_json_schema(model)


_cache = {}


def _to_json_schema(model):
    if model not in _cache:
        _cache[model] = _to_json_schema_no_cache(model)
    return _cache[model]


def _to_json_schema_no_cache(model):
    if isinstance(model, ModelType):
        return _model_type_to_json_schema(model)
    elif isinstance(model, ListType):
        return _list_type_to_json_schema(model)
    elif isinstance(model, DictType):
        return _dict_type_to_json_schema(model)
    else:
        for cls, schema in JSON_SCHEMA_MAP.items():
            if isinstance(model, cls):
                return schema
    if isinstance(model, BaseType):
        return {"type": "object"}
    if isinstance(model, Serializable):
        return _to_json_schema_no_cache(model.type)
    raise SerializationException(
        "unable to create json schema for type "
        + "{0}. Expected a primitive or ".format(model)
        + " a schematics model or type"
    )


def _model_type_to_json_schema(model):
    required = []
    schema = {"title": model.model_name, "type": "object", "properties": {}}
    for name, field in model.fields.items():
        if field.required:
            required.append(name)
        schema["properties"][name] = _to_json_schema(field)
    if len(required) > 0:
        schema["required"] = required
    return schema


def _list_type_to_json_schema(list_type):
    return {"type": "array", "items": _to_json_schema(list_type.field)}


def _dict_type_to_json_schema(dict_type):
    return {"type": "object", "additionalProperties": _to_json_schema(dict_type.field)}


def _enforce_instance(model_or_class):
    """
    It's a common mistake to not initialize a
    schematics class. We should handle that by just
    calling the default constructor.
    """
    if isinstance(model_or_class, type) and issubclass(model_or_class, BaseType):
        return model_or_class()
    return model_or_class



class Template(object):
    """
    a rudimentary replacement for Jinja2. currently in place
    due to a bug with  the typing module.
    """

    def __init__(self, template_string):
        self._template_string = template_string

    def render(self, **parameters):
        chunks = []
        iterator = iter(self._template_string)
        while True:
            try:
                c = next(iterator)
                if c == "{":
                    next_c = next(iterator)
                    if next_c == "{":
                        chunks.append(_capture_variable(iterator, parameters))
                    else:
                        chunks.append(c)
                        chunks.append(next_c)
                    continue
                chunks.append(c)
            except StopIteration:
                break
        return "".join(chunks)


def _capture_variable(iterator, parameters):
    """
    return the replacement string.
    this assumes the preceeding {{ has already been
    popped off.
    """
    key = ""
    next_c = next(iterator)
    while next_c != "}":
        key += next_c
        next_c = next(iterator)
    # remove the final "}"
    next(iterator)
    return parameters[key]



from ..attributes import TransmuteAttributes
from .signature import FunctionSignature
from .response import Response
from .transmute_function import TransmuteFunction


__all__ = ["FunctionSignature", "Response", "TransmuteAttributes", "TransmuteFunction"]



from .function.signature import NoDefault
from transmute_core.http_parameters import ParamSet, Param
from .exceptions import APIException, NoSerializerFound

NoArgument = object()


class ParamExtractor(object):
    """
    An abstract object to extend to handle
    extracting parameters from a request.
    """

    NoArgument = NoArgument

    def extract_params(self, context, transmute_func, content_type):
        parameters = transmute_func.parameters
        signature = transmute_func.signature
        content_type = content_type or context.contenttype_serializers.default.main_type

        args = {}
        framework_args = self._get_framework_args()
        for key, value in framework_args.items():
            if signature.get_argument(key):
                args[key] = value

        empty_args = []

        for name, param in parameters.query.items():
            arginfo = param.arginfo
            if name in framework_args:
                continue
            schema = context.serializers.to_json_schema(arginfo.type)
            is_list = "items" in schema
            values = self._query_argument(name, is_list)
            if values is NoArgument:
                empty_args.append(arginfo)
            else:
                if is_list and len(values) == 1:
                    values = values[0].split(",")
                args[name] = context.serializers.load(arginfo.type, values)

        for name, param in parameters.header.items():
            arginfo = param.arginfo
            if name in framework_args:
                continue
            value = self._header_argument(name)
            if value is NoArgument:
                empty_args.append(arginfo)
            else:
                args[name] = context.serializers.load(arginfo.type, value)

        if isinstance(parameters.body, Param) or len(parameters.body) > 0:
            try:
                serializer = context.contenttype_serializers[content_type]
            except NoSerializerFound:
                raise APIException(
                    "unable to extract parameters for content type {0}. Supported types are: {1}".format(
                        content_type, context.contenttype_serializers.keys()
                    )
                )
            if not self.body:
                body_dict = {}
            else:
                body_dict = serializer.load(self.body)

            if isinstance(parameters.body, Param):
                arginfo = parameters.body.arginfo
                args[arginfo.name] = context.serializers.load(arginfo.type, body_dict)
            else:
                for name, param in parameters.body.items():
                    arginfo = param.arginfo
                    if name in framework_args:
                        continue
                    if name in body_dict:
                        args[name] = context.serializers.load(
                            arginfo.type, body_dict[name]
                        )
                    else:
                        empty_args.append(arginfo)

        for name, param in parameters.path.items():
            arginfo = param.arginfo
            if name in framework_args:
                continue
            values = self._path_argument(name)
            if values is not NoArgument:
                args[name] = context.serializers.load(arginfo.type, values)
            else:
                empty_args.append(arginfo)

        required_params_not_passed = []

        for arg in empty_args:
            if arg.default is NoDefault:
                required_params_not_passed.append(arg.name)
            else:
                args[arg.name] = arg.default

        if len(required_params_not_passed) > 0:
            raise APIException(
                "required arguments {0} not passed".format(required_params_not_passed)
            )

        return signature.split_args(args)

    def _get_framework_args(self):
        """
        often, a framework provides specific variables that are passed
        into the handler function (e.g. the request object in
        aiohttp). return a dictionary of these arguments, which will be
        added to the function arguments if they appear.
        """
        pass

    @property
    def body(self):
        """ return the request body. """
        raise NotImplementedError()

    def _query_argument(self, key, is_list):
        raise NotImplementedError()

    def _header_argument(self, key):
        raise NotImplementedError()

    def _path_argument(self, key):
        raise NotImplementedError()


def _fold(param_or_param_set, value_dict, request_context):
    # a single parameter consumes the whole dict
    if isinstance(param_or_param_set, Param):
        arginfo = param_or_param_set.arginfo
        value = request_context.context.serializers.load(arginfo.type, value_dict)
        request_context.args[argument.name] = value
    # a param set consumes the dict piece by piece
    else:
        for name, param in param_or_param_set.values():
            arginfo = param.arginfo
            if name in request_context.framework_args:
                continue
            if name in value_dict:
                value = context.serializers.load(arginfo.type, value_dict[name])
                request_context.args[arginfo.name] = value
            elif arginfo.default is not NoDefault:
                request_context.args[arginfo.name] = arginfo.default
            else:
                request_context.empty_args.append(name)



import pytest
from transmute_core.http_parameters import Parameters, Param, ParamSet
from transmute_core.function.signature import Argument, NoDefault


@pytest.fixture
def parameters():
    return Parameters(
        query=ParamSet(
            {
                "query": Param(
                    argument_name="query", arginfo=Argument("query", None, int)
                )
            }
        ),
        body=ParamSet(
            {
                "left": Param(
                    argument_name="left", arginfo=Argument("left", NoDefault, int)
                ),
                "right": Param(
                    argument_name="right", arginfo=Argument("right", 2, int)
                ),
            }
        ),
        header=ParamSet(
            {
                "header": Param(
                    argument_name="header", arginfo=Argument("header", NoDefault, int)
                )
            }
        ),
        path=ParamSet(
            {
                "path": Param(
                    argument_name="path", arginfo=Argument("path", NoDefault, int)
                )
            }
        ),
    )


@pytest.fixture
def parameter_post_schema():
    return Parameters(
        body=Param(argument_name="body", arginfo=Argument("left", NoDefault, int))
    )



from transmute_core.function.signature import Argument, NoDefault
from transmute_core.http_parameters import get_swagger_parameters
from transmute_core import default_context
from transmute_core.http_parameters import Parameters, Param, ParamSet


def test_swagger_parameters():
    parameters = Parameters(
        query=ParamSet(
            {
                "query": Param(
                    argument_name="query", arginfo=Argument("query", None, int)
                )
            }
        ),
        body=ParamSet(
            {
                "left": Param(
                    argument_name="left", arginfo=Argument("left", NoDefault, int)
                ),
                "right": Param(
                    argument_name="right", arginfo=Argument("right", 2, int)
                ),
            }
        ),
        header=ParamSet(
            {
                "header": Param(
                    argument_name="header", arginfo=Argument("header", NoDefault, int)
                )
            }
        ),
        path=ParamSet(
            {
                "path": Param(
                    argument_name="path", arginfo=Argument("path", NoDefault, int)
                )
            }
        ),
    )
    params = get_swagger_parameters(parameters, default_context)
    params = [p.to_primitive() for p in params]
    assert {
        "in": "query",
        "name": "query",
        "required": False,
        "type": "integer",
        "collectionFormat": "multi",
        "description": "",
    } in params
    assert {
        "in": "body",
        "name": "body",
        "schema": {
            "type": "object",
            "required": ["left"],
            "properties": {
                "left": {"type": "integer", "description": ""},
                "right": {"type": "integer", "description": ""},
            },
        },
        "required": True,
        "description": "",
    } in params
    assert {
        "in": "header",
        "name": "header",
        "type": "integer",
        "required": True,
        "description": "",
    } in params
    assert {
        "in": "path",
        "name": "path",
        "type": "integer",
        "required": True,
        "description": "",
    } in params



#!/usr/bin/env python
import glob
import os
import sys
from setuptools import setup, find_packages

is_release = False
if "--release" in sys.argv:
    is_release = True
    sys.argv.remove("--release")

base = os.path.dirname(os.path.abspath(__file__))

README_PATH = os.path.join(base, "README.rst")

install_requires = [
    'attrs>=17.3.0',
    'cattrs>=0.9.0',
    'jsonschema-extractor>=0.6.0',
    'schematics>=2.0.0',
    'six',
    'swagger-schema>=0.5.1',
    'pyyaml',
]

if sys.version_info < (3, 5):
    install_requires.append("typing")

if sys.version_info < (3, 4):
    install_requires.append("singledispatch")

tests_require = []

data_files = []

swagger_statics = glob.glob("transmute_core/swagger/static/**/")
for directory in swagger_statics:
    data_files.append(os.path.join("swagger", "static", directory, "*"))

setup(name='transmute-core',
      setup_requires=["vcver"],
      vcver={
          "is_release": is_release,
          "path": base
      },
      description=(
          "a utility library to help provide api route "
          "generation form function signature for web "
          "frameworks."
      ),
      long_description=open(README_PATH).read(),
      author='Yusuke Tsutsumi',
      author_email='yusuke@tsutsumi.io',
      url='https://github.com/toumorokoshi/transmute-core',
      # data_files=data_files,
      package_data={"transmute_core": data_files},
      packages=find_packages(),
      include_package_data=True,
      install_requires=install_requires,
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Operating System :: MacOS',
          'Operating System :: POSIX :: Linux',
          'Topic :: System :: Software Distribution',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
      ],
      tests_require=tests_require
      )



from transmute_core.exceptions import APIException, SerializationException


def test_serialization_error_is_api_exception():
    """
    a serialization exception should
    be considered in the default exceptions
    of the api.
    """
    assert isinstance(SerializationException(""), APIException)



import pytest
from transmute_core import NoSerializerFound, SerializationException


def test_default_serializer_json(serializer_set):
    frm, expected_to = {"foo": "bar"}, b'{"foo": "bar"}'
    serializer = serializer_set["application/json"]
    assert serializer.dump(frm) == expected_to
    assert serializer.load(expected_to) == frm


def test_default_serializer_yaml(serializer_set):
    frm, expected_to = {"foo": "bar"}, b"foo: bar\n"
    serializer = serializer_set["application/yaml"]
    assert serializer.dump(frm) == expected_to
    assert serializer.load(expected_to) == frm


def test_default_serializer_prop(serializer_set):
    assert serializer_set.default.main_type == "application/json"


def test_no_serializer_found_raises_exception(serializer_set):
    with pytest.raises(NoSerializerFound):
        assert serializer_set["oogabooga"]


@pytest.mark.parametrize(
    "content_type,bad_input",
    [("application/yaml", b"[a, !eafia']atedntad}"), ("application/json", b'{"ooga')],
)
def test_bad_object_raises_serialization_exception(
    serializer_set, content_type, bad_input
):
    """ a bad object serialization should raise a serialization exception """
    with pytest.raises(SerializationException):
        serializer_set[content_type].load(bad_input)


def test_keys(serializer_set):
    assert serializer_set.keys() == ["application/json", "application/x-yaml"]



import sys
from transmute_core import process_result, ParamExtractor


class ParamExtractorMock(ParamExtractor):
    def __init__(self, body):
        self._body = body

    @property
    def body(self):
        return self._body

    def _get_framework_args(self):
        return {}

        return NoArgument

    def _header_argument(self, key):
        return NoArgument

    def _path_argument(self, key):
        return NoArgument


def execute(context, func, obj_as_json):
    extractor = ParamExtractorMock(obj_as_json)
    args, kwargs = extractor.extract_params(context, func, "application/json")
    exc, result = None, None
    try:
        result = func(*args, **kwargs)
    except Exception as e:
        exc = e
        exc.__traceback__ = sys.exc_info[:2]
    process_result(func, context, result, exc, "application/json")



from transmute_core.compat import getfullargspec
from transmute_core import annotate
from transmute_core.function.signature import FunctionSignature, NoDefault


def test_signature():
    @annotate({"x": int, "y": float, "width": int, "height": float})
    def make_square(x, y, width=None, height=12):
        pass

    argspec = getfullargspec(make_square)
    signature = FunctionSignature.from_argspec(argspec)
    assert len(signature.args) == 2
    assert signature.args[0].name == "x"
    assert signature.args[0].default == NoDefault
    assert signature.args[0].type == int
    assert len(signature.kwargs) == 2
    assert signature.kwargs["width"].name == "width"
    assert signature.kwargs["width"].default is None
    assert signature.kwargs["width"].type == int


def test_signature_no_kwargs():
    @annotate({"x": int, "y": int, "return": int})
    def make_square(x, y):
        return x * y

    argspec = getfullargspec(make_square)
    signature = FunctionSignature.from_argspec(argspec)
    assert len(signature.args) == 2
    assert signature.args[0].name == "x"
    assert signature.args[0].default == NoDefault
    assert signature.args[0].type == int
    assert signature.args[1].name == "y"
    assert signature.args[1].default == NoDefault
    assert signature.args[1].type == int
    assert len(signature.kwargs) == 0


def test_self_signature():
    def square(self, resource, multiplier=None):
        pass

    argspec = getfullargspec(square)
    signature = FunctionSignature.from_argspec(argspec)

    assert len(signature.args) == 1
    assert len(signature.kwargs) == 1
    assert signature.kwargs["multiplier"].name == "multiplier"


def test_get_signature():
    def square(self, resource, multiplier=None):
        pass

    argspec = getfullargspec(square)
    signature = FunctionSignature.from_argspec(argspec)

    arg = signature.get_argument("resource")
    assert arg.name == "resource"

    kwarg = signature.get_argument("multiplier")
    assert kwarg.name == "multiplier"
